@@ -18,19 +18,18 @@
 //
 //===----------------------------------------------------------------------===//
 #include "polly/ScopDetectionDiagnostic.h"
-
-#include "llvm/Analysis/LoopInfo.h"
+#include "polly/Support/ScopLocation.h"
+#include "llvm/ADT/SmallVector.h"
+#include "llvm/ADT/Statistic.h"
 #include "llvm/Analysis/AliasSetTracker.h"
+#include "llvm/Analysis/LoopInfo.h"
+#include "llvm/Analysis/RegionInfo.h"
 #include "llvm/IR/BasicBlock.h"
 #include "llvm/IR/DebugInfo.h"
 #include "llvm/IR/DebugLoc.h"
 #include "llvm/IR/DiagnosticInfo.h"
 #include "llvm/IR/LLVMContext.h"
 #include "llvm/IR/Value.h"
-#include "llvm/ADT/SmallVector.h"
-#include "llvm/ADT/Statistic.h"
-
-#include "llvm/Analysis/RegionInfo.h"
 
 #define DEBUG_TYPE "polly-detect"
 #include "llvm/Support/Debug.h"
@@ -60,29 +59,6 @@ template <typename T> std::string operator+(Twine LHS, const T &RHS) {
 
   return LHS.concat(Buf).str();
 }
-
-void getDebugLocation(const Region *R, unsigned &LineBegin, unsigned &LineEnd,
-                      std::string &FileName) {
-  LineBegin = -1;
-  LineEnd = 0;
-
-  for (const BasicBlock *BB : R->blocks())
-    for (const Instruction &Inst : *BB) {
-      DebugLoc DL = Inst.getDebugLoc();
-      if (DL.isUnknown())
-        continue;
-
-      DIScope Scope(DL.getScope(Inst.getContext()));
-
-      if (FileName.empty())
-        FileName = Scope.getFilename();
-
-      unsigned NewLine = DL.getLine();
-
-      LineBegin = std::min(LineBegin, NewLine);
-      LineEnd = std::max(LineEnd, NewLine);
-    }
-}
 }
 
 namespace llvm {
@@ -98,11 +74,11 @@ static void getDebugLocations(const Region *R, DebugLoc &Begin, DebugLoc &End) {
   for (const BasicBlock *BB : R->blocks())
     for (const Instruction &Inst : *BB) {
       DebugLoc DL = Inst.getDebugLoc();
-      if (DL.isUnknown())
+      if (!DL)
         continue;
 
-      Begin = Begin.isUnknown() ? DL : std::min(Begin, DL);
-      End = End.isUnknown() ? DL : std::max(End, DL);
+      Begin = Begin ? std::min(Begin, DL) : DL;
+      End = End ? std::max(End, DL) : DL;
     }
 }
 
@@ -119,8 +95,7 @@ void emitRejectionRemarks(const llvm::Function &F, const RejectLog &Log) {
       "The following errors keep this region from being a Scop.");
 
   for (RejectReasonPtr RR : Log) {
-    const DebugLoc &Loc = RR->getDebugLoc();
-    if (!Loc.isUnknown())
+    if (const DebugLoc &Loc = RR->getDebugLoc())
       emitOptimizationRemarkMissed(Ctx, DEBUG_TYPE, F, Loc,
                                    RR->getEndUserMessage());
   }
@@ -232,6 +207,22 @@ bool ReportInvalidCond::classof(const RejectReason *RR) {
   return RR->getKind() == rrkInvalidCond;
 }
 
+//===----------------------------------------------------------------------===//
+// ReportUnsignedCond.
+
+std::string ReportUnsignedCond::getMessage() const {
+  return ("Condition in BB '" + BB->getName()).str() +
+         "' performs a comparision on (not yet supported) unsigned integers.";
+}
+
+std::string ReportUnsignedCond::getEndUserMessage() const {
+  return "Unsupported comparision on unsigned integers encountered";
+}
+
+bool ReportUnsignedCond::classof(const RejectReason *RR) {
+  return RR->getKind() == rrkUnsignedCond;
+}
+
 //===----------------------------------------------------------------------===//
 // ReportUndefOperand.
 
@@ -591,4 +582,30 @@ const DebugLoc &ReportEntry::getDebugLoc() const {
 bool ReportEntry::classof(const RejectReason *RR) {
   return RR->getKind() == rrkEntry;
 }
+
+//===----------------------------------------------------------------------===//
+// ReportUnprofitable.
+ReportUnprofitable::ReportUnprofitable(Region *R)
+    : ReportOther(rrkUnprofitable), R(R) {}
+
+std::string ReportUnprofitable::getMessage() const {
+  return "Region can not profitably be optimized!";
+}
+
+std::string ReportUnprofitable::getEndUserMessage() const {
+  return "No profitable polyhedral optimization found";
+}
+
+const DebugLoc &ReportUnprofitable::getDebugLoc() const {
+  for (const BasicBlock *BB : R->blocks())
+    for (const Instruction &Inst : *BB)
+      if (const DebugLoc &DL = Inst.getDebugLoc())
+        return DL;
+
+  return R->getEntry()->getTerminator()->getDebugLoc();
+}
+
+bool ReportUnprofitable::classof(const RejectReason *RR) {
+  return RR->getKind() == rrkUnprofitable;
+}
 } // namespace polly
