@@ -22,14 +22,14 @@
 #include "polly/RegisterPasses.h"
 #include "polly/Canonicalization.h"
 #include "polly/CodeGen/CodeGeneration.h"
-#include "polly/Dependences.h"
+#include "polly/DependenceInfo.h"
 #include "polly/LinkAllPasses.h"
 #include "polly/Options.h"
 #include "polly/ScopDetection.h"
 #include "polly/ScopInfo.h"
 #include "polly/TempScopInfo.h"
 #include "llvm/Analysis/CFGPrinter.h"
-#include "llvm/PassManager.h"
+#include "llvm/IR/LegacyPassManager.h"
 #include "llvm/Transforms/IPO/PassManagerBuilder.h"
 #include "llvm/Transforms/Scalar.h"
 #include "llvm/Transforms/Vectorize.h"
@@ -44,46 +44,39 @@ static cl::opt<bool>
     PollyEnabled("polly", cl::desc("Enable the polly optimizer (only at -O3)"),
                  cl::init(false), cl::ZeroOrMore, cl::cat(PollyCategory));
 
-enum OptimizerChoice {
-  OPTIMIZER_NONE,
-#ifdef PLUTO_FOUND
-  OPTIMIZER_PLUTO,
-#endif
-  OPTIMIZER_ISL
-};
+static cl::opt<bool> PollyDetectOnly(
+    "polly-only-scop-detection",
+    cl::desc("Only run scop detection, but no other optimizations"),
+    cl::init(false), cl::ZeroOrMore, cl::cat(PollyCategory));
+
+enum OptimizerChoice { OPTIMIZER_NONE, OPTIMIZER_ISL };
 
 static cl::opt<OptimizerChoice> Optimizer(
     "polly-optimizer", cl::desc("Select the scheduling optimizer"),
     cl::values(clEnumValN(OPTIMIZER_NONE, "none", "No optimizer"),
-#ifdef PLUTO_FOUND
-               clEnumValN(OPTIMIZER_PLUTO, "pluto",
-                          "The Pluto scheduling optimizer"),
-#endif
                clEnumValN(OPTIMIZER_ISL, "isl", "The isl scheduling optimizer"),
                clEnumValEnd),
     cl::Hidden, cl::init(OPTIMIZER_ISL), cl::ZeroOrMore,
     cl::cat(PollyCategory));
 
-CodeGenChoice polly::PollyCodeGenChoice;
-static cl::opt<CodeGenChoice, true> XCodeGenerator(
+enum CodeGenChoice { CODEGEN_ISL, CODEGEN_NONE };
+static cl::opt<CodeGenChoice> CodeGenerator(
     "polly-code-generator", cl::desc("Select the code generator"),
     cl::values(clEnumValN(CODEGEN_ISL, "isl", "isl code generator"),
                clEnumValN(CODEGEN_NONE, "none", "no code generation"),
                clEnumValEnd),
-    cl::Hidden, cl::location(PollyCodeGenChoice), cl::init(CODEGEN_ISL),
-    cl::ZeroOrMore, cl::cat(PollyCategory));
+    cl::Hidden, cl::init(CODEGEN_ISL), cl::ZeroOrMore, cl::cat(PollyCategory));
 
 VectorizerChoice polly::PollyVectorizerChoice;
 static cl::opt<polly::VectorizerChoice, true> Vectorizer(
     "polly-vectorizer", cl::desc("Select the vectorization strategy"),
-    cl::values(clEnumValN(polly::VECTORIZER_NONE, "none", "No Vectorization"),
-               clEnumValN(polly::VECTORIZER_POLLY, "polly",
-                          "Polly internal vectorizer"),
-               clEnumValN(polly::VECTORIZER_UNROLL_ONLY, "unroll-only",
-                          "Only grouped unroll the vectorize candidate loops"),
-               clEnumValN(polly::VECTORIZER_BB, "bb",
-                          "The Basic Block vectorizer driven by Polly"),
-               clEnumValEnd),
+    cl::values(
+        clEnumValN(polly::VECTORIZER_NONE, "none", "No Vectorization"),
+        clEnumValN(polly::VECTORIZER_POLLY, "polly",
+                   "Polly internal vectorizer"),
+        clEnumValN(polly::VECTORIZER_STRIPMINE, "stripmine",
+                   "Strip-mine outer loops for the loop-vectorizer to trigger"),
+        clEnumValEnd),
     cl::location(PollyVectorizerChoice), cl::init(polly::VECTORIZER_NONE),
     cl::ZeroOrMore, cl::cat(PollyCategory));
 
@@ -130,19 +123,12 @@ static cl::opt<bool>
                cl::desc("Show the Polly CFG right after code generation"),
                cl::Hidden, cl::init(false), cl::cat(PollyCategory));
 
-bool polly::PollyAnnotateAliasScopes;
-static cl::opt<bool, true> XPollyAnnotateAliasScopes(
-    "polly-annotate-alias-scopes",
-    cl::desc("Annotate memory instructions with alias scopes"),
-    cl::location(PollyAnnotateAliasScopes), cl::init(true), cl::ZeroOrMore,
-    cl::cat(PollyCategory));
-
 namespace polly {
 void initializePollyPasses(PassRegistry &Registry) {
-  initializeIslCodeGenerationPass(Registry);
+  initializeCodeGenerationPass(Registry);
   initializeCodePreparationPass(Registry);
   initializeDeadCodeElimPass(Registry);
-  initializeDependencesPass(Registry);
+  initializeDependenceInfoPass(Registry);
   initializeIndependentBlocksPass(Registry);
   initializeJSONExporterPass(Registry);
   initializeJSONImporterPass(Registry);
@@ -173,18 +159,21 @@ void initializePollyPasses(PassRegistry &Registry) {
 ///
 /// For certain parts of the Polly optimizer, several alternatives are provided:
 ///
-/// As scheduling optimizer we support PLUTO
-/// (http://pluto-compiler.sourceforge.net) as well as the isl scheduling
-/// optimizer (http://freecode.com/projects/isl). The isl optimizer is the
-/// default optimizer.
+/// As scheduling optimizer we support the isl scheduling optimizer
+/// (http://freecode.com/projects/isl).
 /// It is also possible to run Polly with no optimizer. This mode is mainly
 /// provided to analyze the run and compile time changes caused by the
 /// scheduling optimizer.
 ///
 /// Polly supports the isl internal code generator.
-void registerPollyPasses(llvm::PassManagerBase &PM) {
+void registerPollyPasses(llvm::legacy::PassManagerBase &PM) {
   registerCanonicalicationPasses(PM);
 
+  PM.add(polly::createScopDetectionPass());
+
+  if (PollyDetectOnly)
+    return;
+
   PM.add(polly::createScopInfoPass());
 
   if (PollyViewer)
@@ -206,12 +195,6 @@ void registerPollyPasses(llvm::PassManagerBase &PM) {
   case OPTIMIZER_NONE:
     break; /* Do nothing */
 
-#ifdef PLUTO_FOUND
-  case OPTIMIZER_PLUTO:
-    PM.add(polly::createPlutoOptimizerPass());
-    break;
-#endif
-
   case OPTIMIZER_ISL:
     PM.add(polly::createIslScheduleOptimizerPass());
     break;
@@ -220,9 +203,9 @@ void registerPollyPasses(llvm::PassManagerBase &PM) {
   if (ExportJScop)
     PM.add(polly::createJSONExporterPass());
 
-  switch (PollyCodeGenChoice) {
+  switch (CodeGenerator) {
   case CODEGEN_ISL:
-    PM.add(polly::createIslCodeGenerationPass());
+    PM.add(polly::createCodeGenerationPass());
     break;
   case CODEGEN_NONE:
     break;
@@ -245,7 +228,7 @@ static bool shouldEnablePolly() {
 
 static void
 registerPollyEarlyAsPossiblePasses(const llvm::PassManagerBuilder &Builder,
-                                   llvm::PassManagerBase &PM) {
+                                   llvm::legacy::PassManagerBase &PM) {
   if (!polly::shouldEnablePolly())
     return;
 
